//Jose M. Pea
//CompBio, IFM, LiU
//27/07/2007

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <math.h>
#include <time.h>

#include "linalg.h"
#include "normdist.h"

#define MAX_LINE_LENGTH 100000
#define MAX_STATE_LENGTH 100

int n_cases, n_vars, true_net, exact, n_edges;

double alpha, **data;

bool **the_true_net, **the_true_net_u;

char *data_file, **names;

void EE(void);
void IAMB(int version);
void IAMB(int target, int radius);
void interIAMB(int target, bool **adjacencies, double **sigma);
void interIAMB2(int version, int target, bool **adjacencies, double **sigma);
void analysis(bool **adjacencies);
bool path(int x, int y, bool *cond, bool **adjacencies);
bool path2(int x, int y, bool *cond, bool *visited, bool **adjacencies);
void parse_data(void);
double statistic(int var, int target, bool *cond, double **sigma);
double pvalue(double stat, int n_conds);
void covariance(double **sigma);
bool choldc(double **a, int n, double p[]);
double inverse(double **a, int n, int x, int y);
void hpsort(unsigned long n, double ra[], int xa[], int ya[]);

void main(int argc, char* argv[])

{int i, j, k, l, count, min, x, y;
FILE *f;

if(argc!=7 && argc!=9)
	{printf("\n\n Usage: data_file n_cases n_vars true_net(0=no,1=yes) alpha exact(0=no,1=yes) [target radius]. \n\n"); 
	 exit(0);
	}

data_file=argv[1];
n_cases=atoi(argv[2]);
n_vars=atoi(argv[3]);
true_net=atoi(argv[4]);
alpha=atof(argv[5]);
exact=atoi(argv[6]);

srand((unsigned)time(NULL));

data=new double*[n_cases];
for(i=0;i<n_cases;i++)
	data[i]=new double[n_vars];

names=new char*[n_vars];
for(i=0;i<n_vars;i++)
	names[i]=new char[(int)MAX_STATE_LENGTH];

if(true_net==1)
	{the_true_net=new bool*[n_vars];
	 for(i=0;i<n_vars;i++)
		the_true_net[i]=new bool[n_vars];
	}

parse_data();

n_edges=0;
if(true_net==1)
	{the_true_net_u=new bool*[n_vars];
	 for(i=0;i<n_vars;i++)
		{the_true_net_u[i]=new bool[n_vars];
		for(j=0;j<n_vars;j++)
			the_true_net_u[i][j]=0;
		}

	for(i=1;i<n_vars;i++)
		for(j=0;j<i;j++)
			if(the_true_net[i][j]==1 || the_true_net[j][i]==1)
				{the_true_net_u[i][j]=1;
				the_true_net_u[j][i]=1;
				n_edges++;
				}
			else
				for(k=0;k<n_vars;k++)
					if(k!=i && k!=j && the_true_net[i][k]==1 && the_true_net[j][k]==1)
						{the_true_net_u[i][j]=1;
						the_true_net_u[j][i]=1;
						n_edges++;
						}
	}

printf("\n n_vars= %d, n_cases= %d, n_edges= %d",n_vars,n_cases,n_edges);
f=fopen("res.txt","a");
fprintf(f,"\n n_vars= %d, n_cases= %d, n_edges= %d,",n_vars,n_cases,n_edges); 
fclose(f);

if(argc==7)
	{if(n_cases>n_vars)
		EE();
	IAMB(2);
	IAMB(1);
	}
else
	for(i=0;i<n_vars;i++)
		if(strcmp(names[i],argv[7])==0)
			{IAMB(i,atoi(argv[8]));
			break;
			}
}

void EE(void)

{int i, j, n_pvalues, *xs, *ys;
double stat, by, *pvalues, **sigma;
bool **adjacencies;
clock_t clock_init;
FILE *f;

clock_init=clock();

adjacencies=new bool*[n_vars];
for(i=0;i<n_vars;i++)
	{adjacencies[i]=new bool[n_vars];
	for(j=0;j<n_vars;j++)
		adjacencies[i][j]=1;
	adjacencies[i][i]=0;	
	}

sigma=new double*[n_vars];
for(i=0;i<n_vars;i++)
	sigma[i]=new double[n_vars];

covariance(sigma);
inverse(sigma,n_vars,-1,-1);

pvalues=new double[n_vars*(n_vars-1)/2];
xs=new int[n_vars*(n_vars-1)/2];
ys=new int[n_vars*(n_vars-1)/2];

n_pvalues=0;
for(i=1;i<n_vars;i++)
	for(j=0;j<i;j++)
		{if(sigma[i][j]==(double)0)
			stat=(double)0;
		else
			stat=-sigma[i][j]/sqrt(sigma[i][i]*sigma[j][j]);

		if(stat==(double)1)
			stat=(double)INT_MAX;
		else
			{if(exact==0)
				stat=(double)0.5*log(((double)1+stat)/((double)1-stat));
			else
				stat=stat*sqrt((double)(n_cases-2-(n_vars-2))/((double)1-stat*stat));
			}

		pvalues[n_pvalues]=-fabs(stat);
		xs[n_pvalues]=i;
		ys[n_pvalues]=j;
		n_pvalues++;
		}

hpsort((unsigned long)n_pvalues,pvalues-1,xs-1,ys-1);

by=(double)0;
for(i=0;i<n_vars*(n_vars-1)/2;i++)
	by+=(double)1/(double)(i+1);

for(i=n_pvalues-1;i>=0;i--)
	if(pvalue(-pvalues[i],n_vars-2)*by*(double)(n_vars*(n_vars-1)/2)/(double)(i+1)>alpha/(double)2)
		{adjacencies[xs[i]][ys[i]]=0;
		adjacencies[ys[i]][xs[i]]=0;
		}
	else
		break;

printf("\n EE: \t %d sec.,",(int)(clock()-clock_init)/(int)CLOCKS_PER_SEC); 
f=fopen("res.txt","a");
fprintf(f," EE: %d sec.,",(int)(clock()-clock_init)/(int)CLOCKS_PER_SEC); 
fclose(f);
analysis(adjacencies);

for(i=0;i<n_vars;i++)
	delete [] adjacencies[i];
delete [] adjacencies;
for(i=0;i<n_vars;i++)
	delete [] sigma[i];
delete [] sigma;
delete [] pvalues;
delete [] xs;
delete [] ys;
}

void IAMB(int version)
 
{int i, j;
double **sigma;
bool **adjacencies;
clock_t clock_init;
FILE *f;

clock_init=clock();

adjacencies=new bool*[n_vars];
for(i=0;i<n_vars;i++)
	{adjacencies[i]=new bool[n_vars];
	for(j=0;j<n_vars;j++)
		adjacencies[i][j]=0;
	}

sigma=new double*[n_vars];
for(i=0;i<n_vars;i++)
	sigma[i]=new double[n_vars];

covariance(sigma);

for(i=0;i<n_vars;i++)
	if(version==0)	
		interIAMB(i,adjacencies,sigma);
	else
		interIAMB2(version,i,adjacencies,sigma);

printf("\n IAMB%d: \t %d sec.,",version,(int)(clock()-clock_init)/(int)CLOCKS_PER_SEC); 
f=fopen("res.txt","a");
fprintf(f," IAMB%d: %d sec.,",version,(int)(clock()-clock_init)/(int)CLOCKS_PER_SEC); 
fclose(f);
analysis(adjacencies);

for(i=0;i<n_vars;i++)
	delete [] adjacencies[i];
delete [] adjacencies;
for(i=0;i<n_vars;i++)
	delete [] sigma[i];
delete [] sigma;
}

void interIAMB(int target, bool **adjacencies, double **sigma)

//Modified from the original in C:/LearningMBs/IJAR/code/kmb4.cpp.

{int i, n_conds, n_pvalues, *xs;
bool *mb, stop, stop2;
double *pvalues;

pvalues=new double[n_vars];
xs=new int[n_vars];

mb=new bool[n_vars];
for(i=0;i<n_vars;i++)
	mb[i]=0;
n_conds=0;

do
	{stop2=1;
	do
		{n_pvalues=0;
		for(i=0;i<n_vars;i++)
			if(i!=target && mb[i]==0)
				{pvalues[n_pvalues]=-fabs(statistic(i,target,mb,sigma));
				xs[n_pvalues]=i;
				n_pvalues++;
				}

		stop=1;
		if(n_pvalues>0)
			{hpsort((unsigned long)n_pvalues,pvalues-1,xs-1,xs-1);
			
			if(pvalue(-pvalues[0],n_conds)<=alpha/(double)2)
				{mb[xs[0]]=1;
				n_conds++;
				stop=0;
				stop2=0;
				}
			}
		}
	while(stop==0);

	while(stop2==0)
		{n_pvalues=0;
		 for(i=0;i<n_vars;i++)
			if(mb[i]==1)
				{mb[i]=0;
				pvalues[n_pvalues]=-fabs(statistic(i,target,mb,sigma));	
				mb[i]=1;
				xs[n_pvalues]=i;
				n_pvalues++;
				}
					
		stop2=1;
		if(n_pvalues>0)
			{hpsort((unsigned long)n_pvalues,pvalues-1,xs-1,xs-1);

			if(pvalue(-pvalues[n_pvalues-1],n_conds-1)>alpha/(double)2)
				{mb[xs[n_pvalues-1]]=0;
				n_conds--;
				stop2=0;
				stop=0;
				}
			}
		}
	}
while(stop==0);

for(i=0;i<n_vars;i++)
	if(mb[i]==1)
		adjacencies[target][i]=1;

delete [] pvalues;
delete [] xs;
delete [] mb;
}

void interIAMB2(int version, int target, bool **adjacencies, double **sigma)

{int i, n_conds, n_pvalues, *xs;
bool *mb, stop;
double *pvalues, by;

pvalues=new double[n_vars];
xs=new int[n_vars];

mb=new bool[n_vars];
for(i=0;i<n_vars;i++)
	mb[i]=0;
n_conds=0;

by=(double)0;
for(i=0;i<n_vars-1;i++)
	by+=(double)1/(double)(i+1);

do
	{n_pvalues=0;
	 for(i=0;i<n_vars;i++)
		if(i!=target)
			{if(mb[i]==0)
				pvalues[n_pvalues]=pvalue(fabs(statistic(i,target,mb,sigma)),n_conds);
			else
				{mb[i]=0;
				pvalues[n_pvalues]=pvalue(fabs(statistic(i,target,mb,sigma)),n_conds-1);	
				mb[i]=1;
				}
			xs[n_pvalues]=i;
			n_pvalues++;
			}

	hpsort((unsigned long)n_pvalues,pvalues-1,xs-1,xs-1);
	if(version==2)
		for(i=0;i<n_pvalues;i++)
			pvalues[i]*=by*(double)n_pvalues/(double)(i+1);

	stop=1;
	for(i=n_pvalues-1;i>=0;i--)
		if(mb[xs[i]]==1)
			{if(pvalues[i]>alpha/(double)2)
				{mb[xs[i]]=0;
				n_conds--;
				stop=0;
				}
			break;
			}
	if(stop==1)
		for(i=0;i<n_pvalues;i++)
			if(mb[xs[i]]==0)
				{if(pvalues[i]<=alpha/(double)2)
					{mb[xs[i]]=1;
					n_conds++;
					stop=0;
					}
				break;
				}
	}
while(stop==0);

for(i=0;i<n_vars;i++)
	if(mb[i]==1)
		adjacencies[target][i]=1;

delete [] pvalues;
delete [] xs;
delete [] mb;
}

void IAMB(int target, int radius)

{int i, j, k, *tovisit;
double **sigma;
bool **adjacencies;
clock_t clock_init;

clock_init=clock();

adjacencies=new bool*[n_vars];
for(i=0;i<n_vars;i++)
	{adjacencies[i]=new bool[n_vars];
	for(j=0;j<n_vars;j++)
		adjacencies[i][j]=0;
	}

sigma=new double*[n_vars];
for(i=0;i<n_vars;i++)
	sigma[i]=new double[n_vars];

covariance(sigma);

tovisit=new int[n_vars];
for(i=0;i<n_vars;i++)
	tovisit[i]=0;
tovisit[target]=1;

for(k=1;k<=radius;k++)
	for(i=0;i<n_vars;i++)
		if(tovisit[i]==k)
			{interIAMB2(2,i,adjacencies,sigma);

			for(j=0;j<n_vars;j++)
				if(adjacencies[i][j]==1 && tovisit[j]==0)
					tovisit[j]=k+1;
			}


printf("\n IAMB(%d,%d): \t %d sec.,",target,radius,(int)(clock()-clock_init)/(int)CLOCKS_PER_SEC); 
analysis(adjacencies);

delete [] tovisit;
for(i=0;i<n_vars;i++)
	delete [] adjacencies[i];
delete [] adjacencies;
for(i=0;i<n_vars;i++)
	delete [] sigma[i];
delete [] sigma;
}

void analysis(bool **adjacencies)

{int i, j, k, tp, fp, tp2, fp2, n_adjs;
double tpr, fdr, fdr2;
bool *cond, *cond2;
FILE *f;

if(true_net==1)
	{cond=new bool[n_vars];
	cond2=new bool[n_vars];

	tpr=(double)0;
	fdr=(double)0;
	fdr2=(double)0;
	for(i=0;i<n_vars;i++)
		{tp=0;
		fp=0;
		tp2=0;
		fp2=0;
	
		for(j=0;j<n_vars;j++)
			if(adjacencies[i][j]==1)
				{if(the_true_net_u[i][j]==1) 
					tp++;
				else
					{fp++;

					for(k=0;k<n_vars;k++)
						if(k!=j && adjacencies[i][k]==1)
							cond[k]=1;
						else
							cond[k]=0;

					if(path(i,j,cond,the_true_net_u)==0)
						fp2++;
					else
						tp2++;
					}
				}

		tp2+=tp;

		n_adjs=0;
		for(j=0;j<n_vars;j++)
			n_adjs+=(int)the_true_net_u[i][j];

		if(n_adjs>0)
			tpr+=(double)tp/(double)n_adjs;
		if(fp>0)
			fdr+=(double)fp/(double)(tp+fp);
		if(fp2>0)
			fdr2+=(double)fp2/(double)(tp2+fp2);
		}

	printf(" tpr= %.2f, fdr= %.2f, fdr2= %.2f",tpr/(double)n_vars,fdr/(double)n_vars,fdr2/(double)n_vars); 
	f=fopen("res.txt","a");
	fprintf(f," tpr= %.2f, fdr= %.2f, fdr2= %.2f",tpr/(double)n_vars,fdr/(double)n_vars,fdr2/(double)n_vars); 
	fclose(f);

	for(i=0;i<n_vars;i++)
		for(j=0;j<n_vars;j++)
			if(adjacencies[i][j]==1)
				adjacencies[j][i]=1;

	tp=0;
	fp=0;
	tp2=0;
	fp2=0;
	for(i=1;i<n_vars;i++)
		for(j=0;j<i;j++)
			if(adjacencies[i][j]==1)
				{if(the_true_net_u[i][j]==1) 
					tp++;
				else
					{fp++;

					for(k=0;k<n_vars;k++)
						if(k!=j && adjacencies[i][k]==1)
							cond[k]=1;
						else
							cond[k]=0;

					for(k=0;k<n_vars;k++)
						if(k!=i && adjacencies[j][k]==1)
							cond2[k]=1;
						else
							cond2[k]=0;

					if(path(i,j,cond,the_true_net_u)==0 && path(i,j,cond2,the_true_net_u)==0)
						fp2++;
					else
						tp2++;
					}
				}

	tp2+=tp;

	if(n_edges==0)
		tpr=(double)0;
	else
		tpr=tp/(double)n_edges;
	if(fp==0)
		fdr=(double)0;
	else
		fdr=(double)fp/(double)(tp+fp);
	if(fp2==0)
		fdr2=(double)0;
	else
		fdr2=(double)fp2/(double)(tp2+fp2);

	printf(", tpr= %.2f, fdr= %.2f, fdr2= %.2f",tpr,fdr,fdr2); 
	f=fopen("res.txt","a");
	fprintf(f,", tpr= %.2f, fdr= %.2f, fdr2= %.2f",tpr,fdr,fdr2); 
	fclose(f);

	delete [] cond;
	delete [] cond2;
	}
else
	{f=fopen("res.txt","a");
	fprintf(f,"\n\n\n");
	for(i=0;i<n_vars;i++)
		{for(j=0;j<n_vars;j++)
			if(adjacencies[i][j]==1 || adjacencies[j][i]==1)
				fprintf(f,"%d",1);
			else
				fprintf(f,"%d",0);
		fprintf(f,"\n");
		}
	fclose(f);
	}
}

bool path(int x, int y, bool *cond, bool **adjacencies)

{int i;
bool *visited;

visited=new bool[n_vars];
for(i=0;i<n_vars;i++)
	visited[i]=0;

return path2(x,y,cond,visited,adjacencies);

delete [] visited;
}

bool path2(int x, int y, bool *cond, bool *visited, bool **adjacencies)

{int i;

if(x==y)
	return 1;
else
	{visited[x]=1;
	for(i=0;i<n_vars;i++)
		if(adjacencies[x][i]==1 && cond[i]==0 && visited[i]==0)
			if(path2(i,y,cond,visited,adjacencies)==1)
				return 1;
	return 0;
	}
}

void parse_data(void)

{int i, j;
float auxf;
char aux, name2[80], buffer[(int)MAX_LINE_LENGTH], *token;
const char *separators=" ,\t\n\r";
FILE *f_in;

if(true_net==1)
	{sprintf(name2,"%s.net",data_file);
	 f_in=fopen("net.txt","r"); //name2,"r");

	 for(i=0;i<n_vars;i++)
		{for(j=0;j<n_vars;j++)
			{fscanf(f_in,"%c",&aux);
		     the_true_net[i][j]=(bool)atoi(&aux);
			}

		 fscanf(f_in,"%c",&aux);
		}

	 fclose(f_in);
	}

f_in=fopen(data_file,"r");

fgets(buffer,(int)MAX_LINE_LENGTH,f_in);
if((int)strlen(buffer)==(int)MAX_LINE_LENGTH-1)
	{printf("\n\n Exceeding MAX_LINE_LENGTH. \n\n");
	 exit(0);
	}
  
token=strtok(buffer,separators); 
for(j=0;j<n_vars;j++)
	{if((int)strlen(token)>(int)MAX_STATE_LENGTH-1)
		{printf("\n\n Exceeding MAX_STATE_LENGTH. \n\n");
		 exit(0);
		}
		
	 strcpy(names[j],token);
			 
	 token=strtok(0,separators);
	}

for(i=0;i<n_cases;i++)
	for(j=0;j<n_vars;j++)
		{fscanf(f_in,"%f",&auxf);
		 data[i][j]=(double)auxf;
		}

fclose(f_in);
}

double statistic(int var, int target, bool *cond, double **sigma)

//Modification of compute_dep in C:/LearningMBs/IJAR/code/kmb4.cpp.

{int i, j, k, l, n_conds, x, y;
double stat, **sigma2;

n_conds=0;
for(i=0;i<n_vars;i++) 
	n_conds+=(int)cond[i];

sigma2=new double*[n_conds+2];
for(i=0;i<n_conds+2;i++)
	sigma2[i]=new double[n_conds+2];

k=0;
for(i=0;i<n_vars;i++)
	if(i==var || i==target || cond[i]==1)
		{if(i==var)
			x=k;

		l=0;
		for(j=0;j<=i;j++)
			if(j==var || j==target || cond[j]==1)
				{if(j==target)
					y=l;
		
				sigma2[k][l]=sigma[i][j];
				sigma2[l][k]=sigma[j][i];
				l++;
				}
		k++;
		}

stat=inverse(sigma2,n_conds+2,x,y);

if(stat==(double)1)
	stat=(double)INT_MAX;
else
	{if(exact==0)
		stat=(double)0.5*log(((double)1+stat)/((double)1-stat));
	else
		stat=stat*sqrt((double)(n_cases-2-n_conds)/((double)1-stat*stat));
	}

for(i=0;i<n_conds+2;i++)
	delete [] sigma2[i];
delete [] sigma2;

return stat;
}

double pvalue(double stat, int n_conds)

{if(exact==0)
	return (double)1-normalp(stat,(double)0,(double)1/(double)(n_cases-3-n_conds));
else
	return (double)1-studtp(stat,n_cases-2-n_conds);
}

void covariance(double **sigma)

//Unlike in C:/LearningMBs/IJAR/code/kmb4.cpp, here I use the ML estimate though it is biased.
//The Fisher's z and t tests involve this estimate.

{int i, j, k;
double *mean;

mean=new double[n_vars];

for(j=0;j<n_vars;j++)
    {mean[j]=(double)0;
    
	 for(i=0;i<n_cases;i++)
		mean[j]+=data[i][j];

	 mean[j]/=(double)n_cases;
    }

for(j=0;j<n_vars;j++)
	for(k=j;k<n_vars;k++)
		{sigma[j][k]=(double)0;
	
		 for(i=0;i<n_cases;i++)
			sigma[j][k]+=((data[i][j]-mean[j])*(data[i][k]-mean[k]));

		 sigma[j][k]/=(double)n_cases;

		 sigma[k][j]=sigma[j][k];
		}

delete [] mean;
}

bool choldc(double **a, int n, double p[])

//Cholesky decomposition.
//Adapted from Numerical Recipes.
//Note that the algorithm works for one-offset arrays, i.e. a and p must be [1..n][1..n] and [1..n]. See pp. 18-19 in the 1992's book.
//I have adapted the code for zero-offset by substracting 1 from the original index every time an array is accessed.

{int i, j, k;
double sum;

for(i=1;i<=n;i++) 
	{for(j=i;j<=n;j++) 
		{for(sum=a[i-1][j-1],k=i-1;k>=1;k--) 
			sum-=a[i-1][k-1]*a[j-1][k-1];
			
		if(i==j) 
			{if (sum<=(double)0)
				return 0;
			
			p[i-1]=sqrt(sum);
			} 
		else 
			a[j-1][i-1]=sum/p[i-1];
		}
	}

return 1;
}

double inverse(double **a, int n, int x, int y)

//Inversion of a symmetric positive definite matrix.
//Adapted from Numerical Recipes. See p. 98 in the 1992's book.
//Note that the algorithm in p. 98 works for one-offset arrays, i.e. a and p must be [1..n][1..n] and [1..n]. See pp. 18-19 in the 1992's book.
//I have adapted the code for zero-offset by substracting 1 from the original index every time an array is accessed.

{int i, j, k;
double *p, sum, **b, **c, sigma_xy, sigma_xx, sigma_yy;

p=new double[n];

if(choldc(a,n,p)==0)
	{printf("\n\n No Cholesky decomposition (%d). \n\n",n);
	 exit(0);
	}

for(i=1;i<=n;i++)
	{a[i-1][i-1]=(double)1/p[i-1];
	for(j=i+1;j<=n;j++) 
		{sum=(double)0;
		for(k=i;k<j;k++) 
			sum-=a[j-1][k-1]*a[k-1][i-1];
		a[j-1][i-1]=sum/p[j-1];
		}
	}

if(-1<x && x<n_vars && -1<y && y<n_vars)
	{sigma_xy=(double)0;
	sigma_xx=(double)0;
	sigma_yy=(double)0;
	for(i=__max(x,y);i<n;i++)
		sigma_xy+=a[i][x]*a[i][y];
	for(i=x;i<n;i++)
		sigma_xx+=a[i][x]*a[i][x];
	for(i=y;i<n;i++)
		sigma_yy+=a[i][y]*a[i][y];

	delete [] p;

	if(sigma_xy==(double)0)
		return (double)0;
	else
		return -sigma_xy/sqrt(sigma_xx*sigma_yy);
	}
else
	{for(i=0;i<n;i++)
		for(j=i+1;j<n;j++)
			a[i][j]=(double)0;
		
	b=new double*[n];
	for(i=0;i<n;i++)
		b[i]=new double[n];

	c=new double*[n];
	for(i=0;i<n;i++)
		c[i]=new double[n];

	transpose(a,n,n,b);
	matmat(b,n,n,a,n,c);

	for(i=0;i<n;i++)
		for(j=0;j<n;j++)
			a[i][j]=c[i][j];

	delete [] p;
	for(i=0;i<n;i++)
		delete [] b[i];
	delete [] b;
	for(i=0;i<n;i++)
		delete [] c[i];
	delete [] c;

	return (double)1;
	}
}

void hpsort(unsigned long n, double ra[], int xa[], int ya[])

//Heap sort.
//Adapted from Numerical Recipes.
//Note that the algorithm works for one-offset arrays, i.e. ra, xa and ya must be [1..n].
//If your arrays are [0..n-1], then it should be called with ra-1, xa-1 and ya-1. See pp. 18-19 in the 1992's book.

{unsigned long i, ir, j, l;
double rra;
int xra, yra;

if(n<2) 
	return;
l=(n>>1)+1;
ir=n;
for(;;)
	{if(l>1)
		{rra=ra[--l];
		xra=xa[l];
		yra=ya[l];
		}
	else
		{rra=ra[ir];
		xra=xa[ir];
		yra=ya[ir];
		ra[ir]=ra[1];
		xa[ir]=xa[1];
		ya[ir]=ya[1];
		if(--ir==1)
			{ra[1]=rra;
			xa[1]=xra;
			ya[1]=yra;
			break;
			}
		}
	i=l;
	j=l+l;
	while(j<=ir)
		{if(j<ir && ra[j]<ra[j+1])
			j++;
		if(rra<ra[j])
			{ra[i]=ra[j];
			xa[i]=xa[j];
			ya[i]=ya[j];
			i=j;
			j<<=1;
			}
		else
			break;
		}
	ra[i]=rra;
	xa[i]=xra;
	ya[i]=yra;
	}
}